import numpy as np
import posixshmem


def parse_shmem_handle(handle_string):
    buffer_name, dtype, shape, index = handle_string.split("$")
    dtype = getattr(np, dtype)
    shape = tuple(int(n) for n in shape.split(","))
    index = int(index)
    return buffer_name, dtype, shape, index


def open_shmem_from_handle(handle_string):
    """Conveniently open readonly SharedMemory with ndarray view from a handle."""
    buffer_name, dtype, shape, _ = parse_shmem_handle(handle_string)
    buffer_mem = posixshmem.SharedMemory(name=buffer_name, rw=False)

    array = buffer_mem.ndarray(
        shape=shape,
        dtype=dtype,
    )

    return buffer_mem, array


class ShmemCircularBuffer:
    """Convenience wrapper around posixshmem-backed ndarray buffers

    The underlying memory will be opened as an ndarray with shape (buffer_size, ) +
    array_shape where buffer_size is memory_budget // dtype * array size. Each call
    to next_slot will return the next entry along the first dimension of this array
    (both a handle for IPC usage and the ndarray view).
    """

    def __init__(self, memory_budget, array_shape, dtype, shmem_name):
        # for portable use: name has leading slash and no other slashes
        self.shmem_name = "/" + shmem_name.lstrip("/").replace("/", "_")
        self._shared_memory = posixshmem.SharedMemory(
            name=self.shmem_name,
            size=memory_budget,
            rw=True,
        )
        self._buffer_ary = None
        self._update_shape(array_shape, dtype)

    def _update_shape(self, array_shape, dtype):
        array_shape = tuple(array_shape)
        array_bytes = np.dtype(dtype).itemsize * np.product(array_shape)
        num_slots = self._shared_memory.size // array_bytes
        if num_slots == 0:
            raise ValueError("Array size exceeds size of allocated memory block")
        full_shape = (num_slots,) + array_shape

        if self._buffer_ary is not None:
            del self._buffer_ary
        self._buffer_ary = self._shared_memory.ndarray(
            shape=full_shape,
            dtype=dtype,
        )
        shape_str = ",".join(str(n) for n in full_shape)
        self.shmem_handle_template = (
            f"{self.shmem_name}${np.dtype(dtype)}${shape_str}${{index}}"
        )
        self._next_slot_index = 0

    def change_shape(self, array_shape, dtype=None):
        """Set new array shape to buffer. Note that the existing SharedMemory object is
        still used. Old data in there will be mangled and number of slots will depend
        upon new array shape and original memory budget.
        """
        if dtype is None:
            dtype = self._buffer_ary.dtype
        self._update_shape(array_shape, dtype)

    @property
    def num_slots(self):
        return self._buffer_ary.shape[0]

    def next_slot(self):
        current_index = self._next_slot_index
        self._next_slot_index = (self._next_slot_index + 1) % self.num_slots
        shmem_handle = self.shmem_handle_template.format(index=current_index)
        data = self._buffer_ary[current_index]
        return shmem_handle, data
