import collections
import functools
import inspect
import threading
import time
import timeit

import numpy as np


def threadsafe_cache(fun):
    """This decorator imitates functools.cache, but threadsafer

    With multiple threads hitting a function cached by functools.cache, it is possible
    to trigger recomputation. This decorator adds granular locking: each key in the
    cache (derived from arguments) has its own lock.
    """

    locks = {}
    results = {}
    fun_sig = inspect.signature(fun)

    @functools.wraps(fun)
    def aux(*args, **kwargs):
        bound_args = fun_sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        key = bound_args.args + tuple(bound_args.kwargs.items())
        if key in results:
            return results[key]
        with locks.setdefault(key, threading.Lock()):
            if key in results:
                # someone else did this - may still be processing
                return results[key]
            else:
                res = fun(*args, **kwargs)
                results[key] = res
                return res

    return aux


_np_typechar_to_c_typestring = {
    "?": "bool",
    "B": "unsigned char",
    "D": "double complex",
    "F": "float complex",
    "G": "long double complex",
    "H": "unsigned short",
    "I": "unsigned int",
    "L": "unsigned long",
    "Q": "unsigned long long",
    "b": "char",
    "d": "double",
    "e": "half",  # warning: only in CUDA with special support
    "f": "float",
    "g": "long double",
    "h": "short",
    "i": "int",
    "l": "long",
    "q": "long long",
}


def np_dtype_to_c_type(dtype):
    as_char = np.sctype2char(dtype)
    return _np_typechar_to_c_typestring[as_char]


def enum_to_c_template(enum_class):
    res = [f"enum {enum_class.__name__} {{"]
    for field in enum_class:
        res.append(f"\t{field.name} = {field.value},")
    res.append("};")
    return "\n".join(res)


def ceil_div(num, denom):
    return (num + denom - 1) // denom


def shape_after_transpose(input_shape, transpose_pattern, squeeze=True):
    if squeeze:
        input_shape = tuple(dim for dim in input_shape if dim > 1)
    if transpose_pattern is None:
        return input_shape
    return tuple(np.array(input_shape)[list(transpose_pattern)].tolist())


class RepeatingTimer:
    """A timer which will call callback every interval seconds"""

    def __init__(self, interval, callback, timer=timeit.default_timer, start_now=True):
        self.timer = timer
        self.stopped = True
        self.interval = interval
        self.callback = callback
        if start_now:
            self.start()

    def start(self):
        self.stopped = False
        self.wakeup_time = self.timer() + self.interval

        def runner():
            while not self.stopped:
                now = self.timer()
                while now < self.wakeup_time:
                    diff = self.wakeup_time - now
                    time.sleep(diff)
                    if self.stopped:
                        return
                    now = self.timer()
                self.callback()
                self.wakeup_time = self.timer() + self.interval

        self.thread = threading.Thread(target=runner)
        self.thread.start()

    def stop(self):
        self.stopped = True


class ExponentialMovingAverage:
    def __init__(self, alpha, use_first_value=True):
        self.alpha = alpha
        self.initialised = not use_first_value
        self.mean = 0

    def update(self, value):
        if self.initialised:
            self.mean += self.alpha * (value - self.mean)
        else:
            self.mean = value
            self.initialised = True

    def get(self):
        return self.mean


class WindowRateTracker:
    def __init__(self, buffer_size=20, time_window=20, timer=timeit.default_timer):
        self.time_window = time_window
        self.buffer_size = buffer_size
        self.deque = collections.deque(maxlen=self.buffer_size)
        self.timer = timer

    def update(self):
        now = self.timer()
        self.deque.append(now)

    def get(self):
        now = self.timer()
        cutoff = now - self.time_window
        try:
            while self.deque[0] < cutoff:
                self.deque.popleft()
        except IndexError:
            return 0
        if len(self.deque) < self.buffer_size:
            return len(self.deque) / self.time_window
        else:
            # if going faster than buffer size per time window, look at timestamps
            return len(self.deque) / (self.deque[-1] - self.deque[0])
